#!/usr/bin/env python3

''' Command line parsing for zcc'''

import argparse
from logging import debug
from os import system
import sys

from .controller import ControlPoint, ControlPointError
from .single import ControlPointSingle, ControlPointSingleError


def __connect(options: object) -> ControlPoint:

    if options.host and options.port:
        return ControlPoint(options.host, options.port,
                            timeout=options.timeout,
                            debug=options.debug,
                            verbose=options.verbose)
    else:
        return ControlPoint(timeout=options.timeout,
                            debug=options.debug,
                            verbose=options.verbose)


def __connect_single(options: object) -> ControlPointSingle:

    if options.host and options.port:
        return ControlPointSingle(options.host, options.port, device=options.device,
                                  timeout=options.timeout, debug=options.debug)
    else:
        raise ControlPointSingleError('No valid host and/or port given')


def __options(args):

    parser = argparse.ArgumentParser(prog='zcc',
                                     description='''
                                     A command line interface to a ZIMI home controller.
                                     
                                     Operates in two basic modes: (a) discovery mode where 
                                     the local network is scanned for a ZCC and the details 
                                     are printed, or (b) an execute mode where an action is 
                                     sent to a device''')

    parser.add_argument("--debug", action="store_true",
                        help="enable logging")
    parser.add_argument("--verbose", action="store_true",
                        help="enable verbose logging")

    command_group = parser.add_argument_group('command')

    cxg = command_group.add_mutually_exclusive_group(required=True)
    cxg.add_argument("--discover", action="store_true",
                     help="discover ZCC and print devices")
    cxg.add_argument("--execute", action="store_true",
                                         help="execute an action on a device")

    command_group.add_argument("--quick", action="store_true",
                               help="perform a quick action (no zcc device discovery)")

    host_group = parser.add_argument_group('host')
    host_group.add_argument("--host", action="store",
                            help="zcc host name|address")
    host_group.add_argument("--port", action="store", type=int,
                            help="zcc port")
    host_group.add_argument("--timeout", action="store", type=int, default=3,
                            help="zcc timeout value")

    device_group = parser.add_argument_group('device')
    device_group.add_argument("--device", action="store",
                              help="device identifier")
    device_group.add_argument("--action", action="store", type=str,
                              choices=["CloseDoor", "OpenDoor", "TurnOn", "TurnOff",
                                       "OpenToPercentage", "SetBrightness", "SetFanSpeed"])
    device_group.add_argument("--value", action="store",
                              help="device action value (for actions that require)")

    return parser.parse_args(args)


def __main(args):

    options = __options(args)

    if options.discover:

        controller = __connect(options)
        controller.print_description()

    if options.execute:

        try:

            if options.action and options.device:

                if options.quick:
                    controller = __connect_single(options)
                    device = controller
                else:
                    controller = __connect(options)
                    device = controller.devices.get(options.device, None)

                if device:
                    if 'CloseDoor' in options.action:
                        device.close_door()
                    elif 'OpenDoor' in options.action:
                        device.open_door()
                    elif 'TurnOn' in options.action:
                        device.turn_on()
                    elif 'TurnOff' in options.action:
                        device.turn_off()
                    elif 'OpenToPercentage' in options.action:
                        if options.value:
                            device.open_to_percentage(options.value)
                        else:
                            raise ControlPointError(
                                'OpenToPercentage requires --value')
                    elif 'SetBrightness' in options.action:
                        if options.value:
                            device.set_brightness(options.value)
                        else:
                            raise ControlPointError(
                                'SetBrightness requires --value')
                    elif 'SetFanSpeed' in options.action:
                        if options.value:
                            device.set_fanspeed(options.value)
                        else:
                            raise ControlPointError(
                                'SetFanSpeed requires --value')
                    else:
                        raise ControlPointError(
                            'No valid --device and --action')
                else:
                    raise ControlPointError(
                        'No device %s exists' % options.device)
            else:
                raise ControlPointError(
                    '--execute needs --device and --action options')

        except ControlPointError as e:
            print(e)
            sys.exit(1)

    if controller:
        controller.disconnect()


if __name__ == '__main__':

    __main(sys.argv[1:])
