'''ZCC Controller Class.'''
from __future__ import annotations

import json
from json.decoder import JSONDecodeError
from pprint import pformat
import logging
import socket
import sys
import threading
import time
from typing import Dict, List

from zcc.device import ControlPointDevice
from zcc.socket import ControlPointSocket, ControlPointSocketException
from zcc.protocol import ControlPointProtocol


class ControlPointError(Exception):
    '''Represents a ZCC controller error.'''


class ControlPoint:
    '''Represents the ZCC controller which connects to individual devices'''

    def __init__(self, host: str = None, port: int = None, timeout: int = 3,
                 debug: bool = False, verbose: bool = False):

        self.debug = debug
        self.verbose = verbose

        self.logger = logging.getLogger('ControlPoint')
        if self.debug and self.verbose:
            logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
        elif self.debug:
            logging.basicConfig(stream=sys.stderr, level=logging.INFO)
        else:
            logging.basicConfig(stream=sys.stderr, level=logging.WARNING)

        self.host = host
        self.port = int(port) if port else None
        self.brand = None
        self.product = None
        self.mac = None
        self.available_tcps = None
        self.devices: Dict[ControlPointDevice] = {}
        self.timeout = timeout
        self.found = False
        self.ready = False
        self.update_socket = None

        self.discover_timed_out = False

        if not self.host:
            self.__discover()
        if self.host:
            self.__connect()
        else:
            self.logger.error(
                '__init() failed - unable to discover and connect to ZCC')
            raise ControlPointError(
                '__init() failed - unable to discover and connect to ZCC')

    def __del__(self):
        self.__unsubscribe()

    def __str__(self):
        return pformat(vars(self)) + '\n'

    def __connect(self):
        '''Connect to ZCC, build device table and subscribe to updates'''
        self.logger.info('__connect() starting')

        sync_complete = False
        sync_count = 0

        while not sync_complete:

            sync_count += 1

            (properties_new, properties_updates) = self.__sync('properties')
            (states_new, states_updates) = self.__sync('states')
            (actions_new, actions_updates) = self.__sync('actions')

            if (properties_new + properties_updates) == 0:
                continue
            if (states_new + states_updates) < (properties_new + properties_updates):
                continue
            if (actions_new + actions_updates) < (properties_new + properties_updates):
                continue

            if sync_count > 10:
                self.logger.error(
                    '__sync() failed - unable to sync with ZCC')
                raise ControlPointError(
                    '__sync() failed - unable to to sync with ZCC')

            sync_complete = True

        self.ready = True

        self.update_thread = threading.Thread(
            target=self.__update_background_thread)
        self.update_thread.daemon = True
        self.update_lock = threading.Lock()

        self.update_thread.start()

        self.logger.info(
            '__connect() success - connected to ZCC at %s %s', self.host, str(self.port))

    def __discover(self):
        '''Discover ZCC host via UDP broadcast'''

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.settimeout(self.timeout)

        server_address = ('255.255.255.255',
                          ControlPointProtocol.UDP_SEND_PORT)
        message = ControlPointProtocol.UDP_SEND_MSG

        self.discover_timed_out = False

        self.discover_lock = threading.Lock()
        self.discover_thread = threading.Thread(
            target=self.__discover_background_thread)
        self.discover_thread.daemon = True
        self.discover_thread.start()

        self.discover_background_timeout = threading.Timer(
            10.0, self.__discover_background_timeout)
        self.discover_background_timeout.start()

        while not self.found and not self.discover_timed_out:
            sock.sendto(message.encode(), server_address)
            self.logger.info('__discover() sending discover message')
            time.sleep(1)

        if self.host:
            self.logger.info(
                '__discover() success - discovered ZCC at %s %s', self.host, str(self.port))

    def __discover_background_thread(self):
        '''Worker thread to monitor UDP broadcast for ZCC'''
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        server_address = ('', ControlPointProtocol.UDP_RECV_PORT)
        sock.bind(server_address)

        self.logger.info('__discover_background_thread() listening')

        while not self.found and not self.discover_timed_out:
            data, address = sock.recvfrom(4096)
            data = str(data.decode('UTF-8'))
            self.logger.info(
                '__discover_background_thread() rcvd %s from %s', data, str(address))
            lines = data.split('\n')
            for line in lines:
                try:
                    response = json.loads(line)
                    if response:
                        self.brand = response['brand']
                        self.product = response['product']
                        self.mac = response['mac']
                        self.host = address[0]
                        self.port = response['tcp']
                        self.available_tcps = response['availableTcps']
                        with self.discover_lock:
                            self.found = True
                            return
                except JSONDecodeError:
                    break

    def __discover_background_timeout(self):
        with self.discover_lock:
            if not self.found:
                self.logger.error(
                    '__discover_background_timeout() timeout occured before discovery completed')
            self.discover_timed_out = True

    def __get(self, request_type: str):
        '''Sends a request of type actions|properties|states for target and yields objects'''

        try:
            sock = ControlPointSocket(
                self.host, self.port, timeout=self.timeout)
        except ControlPointSocketException as error:
            self.logger.error(
                '__get() ControlPointSocketException %sd from %s %s',
                error, self.host, str(self.port))
            raise ControlPointError(
                '__get() ControlPointError %s to %s port %s' %
                (error, self.host, self.port)) from error
        except ConnectionRefusedError as error:
            self.logger.error(
                '__get() ConnectionRefusedError %s from %s %s', error, self.host, str(self.port))
            return
        sock.sendall(ControlPointProtocol.get(request_type))
        data = sock.recvall()
        sock.close()

        lines = data.split('\n')
        for line in lines:
            try:
                response = json.loads(line).get('response')
                if response:
                    target_response = response.get(
                        ControlPointProtocol.get_response_key(request_type))
                    if target_response:
                        yield target_response
            except JSONDecodeError:
                break

    def __set(self, identifier: str, action: str, params: object = None):
        '''Sends an action for a device and yields response objects'''

        sock = ControlPointSocket(self.host, self.port, timeout=self.timeout)
        sock.sendall(ControlPointProtocol.set(identifier, action, params))
        self.logger.info("__set() sent %s request to %s", action, identifier)

        for line in sock.recvline(single=True):
            try:
                response = json.loads(line).get('response')
                if response:
                    target_response = response.get(
                        ControlPointProtocol.set_response_key())
                    if target_response:
                        self.logger.info(
                            "__set() rcvd response\n%s", pformat(target_response))
                        yield target_response
            except JSONDecodeError:
                break

        sock.close()

    def __subscribe(self):
        '''Sends a subscribe action and yields response objects'''

        self.update_socket = ControlPointSocket(self.host, self.port)
        self.update_socket.sendall(ControlPointProtocol.subscribe())

        for line in self.update_socket.recvline():
            try:
                response = json.loads(line).get('response')
                if response:
                    target_response = response.get(
                        ControlPointProtocol.subscribe_response_key())
                    if target_response:
                        yield target_response
            except JSONDecodeError:
                break

        self.update_socket.close()

    def __sync(self, category: str):
        '''Connect to ZCC and sync the status of category.
        Return a tuple with (new_records, updated_records).'''
        new_records = 0
        updated_records = 0
        for item in self.__get(category):
            for device in item:
                identifier = device['id']
                if not self.devices.get(identifier):
                    self.devices[identifier] = ControlPointDevice(
                        self, identifier)
                    new_records += 1
                else:
                    updated_records += 1
                if category == 'actions':
                    self.devices[identifier].actions = device[category]
                if category == 'properties':
                    self.devices[identifier].properties = device[category]
                if category == 'states':
                    self.devices[identifier].states = device[category]
        self.logger.info(
            '__connect() received %d new and %d updated records for %s',
            new_records, updated_records, category)

        return (new_records, updated_records)

    def __unsubscribe(self):
        '''Sends an unsubscribe action on existing subscribe socket'''

        if self.update_socket:
            self.update_socket.sendall(ControlPointProtocol.unsubscribe())

    def __update_background_thread(self):
        '''Worker thread to monitor state change events and update'''
        self.logger.info('__connect_background_thread() starting')
        while True:
            for item in self.__subscribe():
                for device in item:
                    identifier = device['id']
                    if not self.devices.get(identifier):
                        self.devices[identifier] = ControlPointDevice(
                            self, identifier)
                    if self.devices[identifier].states != device['states']:
                        with self.update_lock:
                            self.logger.info(
                                '__update_background_thread() updating state for %s to:\n%s',
                                identifier, pformat(device['states']))
                            self.devices[identifier].states = device['states']
                            self.devices[identifier].notify_observers()

    def action(self, identifier: str, action: str, params: object = None):
        '''Sends an action for a device and checks for response in first item.
        After receiving first item the method returns immediately rather than
        wait for subsequent items.'''
        for item in self.__set(identifier, action, params=params):
            for device in item:
                identifier = device['id']
                if self.devices[identifier].states != device['states']:
                    with self.update_lock:
                        self.logger.info(
                            'action() updating state for %s to:\n%s',
                            identifier, pformat(device['states']))
                        self.devices[identifier].states = device['states']
                        self.devices[identifier].notify_observers()
                        return

    def describe(self) -> str:
        '''Return a string representation of ZCC including devices'''
        header = '+' + '-' * 130 + '+'
        if self.host:
            description = header + '\n'
            description += '| ControlPoint: %12s        %8s %8s %27s devices        %16s:%-6d %s Tcps    |\n' % (
                self.mac if self.mac else "n/a",
                self.product if self.product else 'n/a',
                self.brand if self.brand else 'n/a',
                str(len(self.devices)),
                self.host,
                self.port,
                str(self.available_tcps) if self.available_tcps else "n/a")
            description += header + '\n'
            for key in self.devices:
                description += self.devices[key].describe()
            return description
        else:
            return 'ControlPoint: not found'

    def print_description(self):
        '''Print description of the ZCC controller.'''
        print(self.describe())

    def disconnect(self):
        '''Disocnnect from zimi controller'''
        self.__unsubscribe()

    @property
    def doors(self) -> List[ControlPointDevice]:
        '''Return an array with all doors'''
        return list(filter(lambda device: device.type == 'garagedoor', self.devices.values()))

    @property
    def fans(self) -> List[ControlPointDevice]:
        '''Return an array with all fans'''
        return list(filter(lambda device: device.type == 'fan', self.devices.values()))

    @property
    def lights(self) -> List[ControlPointDevice]:
        '''Return an array with all lights (i.e. switch or dimmer type)'''
        return list(filter(lambda device:
                           device.type == 'light' or
                           device.type == 'switch' or
                           device.type == 'dimmer', self.devices.values()))

    @property
    def outlets(self) -> List[ControlPointDevice]:
        '''Return an array with all outlets'''
        return list(filter(lambda device: device.type == 'outlet', self.devices.values()))
