import datetime
import logging
import socket
import typing


class ControlPointSocketException(Exception):
    pass


class ControlPointSocket:
    '''A TCP/IP socket that includes recvall method with timeout'''

    def __init__(self, host: str, port: int, timeout: int = None):
        '''Create new TCP/IP socket to host and port with optional timeout'''
        self.host = host
        self.port = port
        self.logger = logging.getLogger('ControlPointSocket')
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if self.sock and timeout:
            self.sock.settimeout(timeout)

        self.timestamp = datetime.datetime.now()

        try:
            self.sock.connect((self.host, self.port))
        except socket.error as e:
            raise ControlPointSocketException(e)

    def __delta_timestamp(self) -> str:
        '''Return a time delta string'''
        while True:
            timestamp = datetime.datetime.now()
            delta_timestamp = str(timestamp - self.timestamp)[5:10]
            self.timestamp = timestamp
            return delta_timestamp

    def close(self) -> None:
        '''Close socket'''
        self.sock.close()
        self.sock = None

    # def recvline(self):
    #     '''Receive data and yeild it line by line'''
    #     buffer = self.sock.recv(4096).decode()
    #     buffering = True
    #     while buffering:
    #         if "\n" in buffer:
    #             (line, buffer) = buffer.split("\n", 1)
    #             self.logger.debug('recvline() ->\n%s', line)
    #             yield line + "\n"
    #         else:
    #             more = None
    #             try:
    #                 more = self.sock.recv(4096).decode()
    #             except socket.error as error:
    #                 self.logger.info("recvline() caught %s", error)
    #             if not more:
    #                 buffering = False
    #             else:
    #                 buffer += more
    #     if buffer:
    #         self.logger.debug('recvline() ->\n%s', buffer)
    #         yield buffer

    def recvline(self, single=False):
        '''Receive data and yeild it line by line.'''
        buffering = True
        sockFile = self.sock.makefile()
        while buffering:
            try:
                line = sockFile.readline()
                if single:
                    buffering = False
            except socket.error as error:
                buffering = False
                self.logger.info("recvline() caught %s", error)
            self.logger.debug('recvline() ->\n%s', line)
            yield line

    def recvall(self, eof: bytes = None) -> str:
        '''Receive all data returned on the socket until timeout or end of file'''
        buff = b''
        chunk = None
        chunks = 'recvall() '
        while chunk != b'':
            try:
                chunk = self.sock.recv(4096)
                chunks += f'{len(chunk)}@{self.__delta_timestamp()}, '
                buff += chunk
            except socket.timeout:
                chunks += f'timeout@{self.__delta_timestamp()}'
                break
            if eof and eof in buff:
                chunks += f'eof@{self.__delta_timestamp()}'
                break
        self.logger.debug(chunks)
        data = buff.decode()
        self.logger.debug('recvall() ->\n%s', data)
        return data

    def sendall(self, message: str):
        '''Send all bytes of string to socket end-point'''
        self.logger.debug('sendall(%s)', message)
        self.sock.sendall(message.encode())
